% Scut clustering implementation
% 
% Mikko Malinen 2014, 2020, 2024
%
% Maintains cost n*TSE for each cluster
% That is the same as pairwise squared within cluster distances
%
% Takes one point and trials to change cluster. Takes new points until no
% more changes improve the cost function.
% Delta COST of removal and addition is calculated by update formula.
%
% X is the data, k is the number of clusters, ground_truth is the
% filename of ground truth centroids. Try f.eg.
% X = load("s1.txt");
% P = All_pairwise(X,15,10,"s1-cb.txt");
%
% If you do not have the ground truth of your dataset, then write f.eg.
% X = load("mydataset.txt");
% P = All_pairwise(X,15,10,'');

function [bestCostScutPartition] = All_pairwise(X,k,REPEATS,ground_truth) 

%clear all;



tic;

%oldpath = path;
%path(oldpath,'..\..\validity_indices\');


CALCULATE_EPSILON = 0;   % 0=no 1=yes  calculate W_total and epsilon
SAVEPARTITION = 0;    % 0=no 1=yes


% n is automatically set later
%k = 8; % k comes as an argument


Dataset_name = '';
%X = load('..\..\datasets\s1.txt'); k=15; Dataset_name = 's1';
%ground_truth = '../../datasets/ground_truths/s-sets/s1-cb.txt';
%X = load('..\..\datasets\s2.txt'); k=15; Dataset_name = 's2';
%ground_truth = '../../datasets/ground_truths/s-sets/s2-cb.txt';
%X = load('..\..\datasets\s3.txt'); k=15; Dataset_name = 's3';
%ground_truth = '../../datasets/ground_truths/s-sets/s3-cb.txt';
%X = load('..\..\datasets\s4.txt'); k=15; Dataset_name = 's4';
%ground_truth = '../../datasets/ground_truths/s-sets/s4-cb.txt';
%X = load('..\..\datasets\iris.txt');  % 150 points, k=3, d=4
%X = load('X:\clustering\datasets\subsets\iris_subset_50.txt'); % 50 points, k=3, d= 4
%X = load('X:\clustering\datasets\subsets\thyroid_subset_50.txt'); % 50
%points , k= 2, d= 5
%X = load('X:\clustering\datasets\subsets\wine_subset_50.txt'); % 50
%points, k = 3, d = 13
%X = load('X:\clustering\datasets\subsets\breast_subset_50.txt'); % 50
%points, k= 2, d=9
%X = load('X:\clustering\datasets\subsets\s1_subset_150.txt');  % 150 points, k=15, d=2
%X = load('X:\clustering\datasets\subsets\s2_subset_150.txt');  % 150 points, k=15, d=2
%X = load('X:\clustering\datasets\subsets\s3_subset_150.txt');  % 150 points, k=15, d=2
%X = load('X:\clustering\datasets\subsets\s4_subset_150.txt');  % 150 points, k=15, d=2
%
%X = load('X:\clustering\datasets\subsets\s1_subset_50.txt');  % 50 points, k=15, d=2
%X = load('X:\clustering\datasets\subsets\s2_subset_50.txt');  % 50 points, k=15, d=2
%X = load('X:\clustering\datasets\subsets\s3_subset_50.txt');  % 50 points, k=15, d=2
%X = load('X:\clustering\datasets\subsets\s4_subset_50.txt');  % 50 points, k=15, d=2
%X = load('X:\clustering\datasets\subsets\a1_subset_50.txt'); % 50 k=20
%X = load('X:\clustering\datasets\subsets\DIM032_subset_50.txt'); % 50 k=16
%X = load('X:\clustering\datasets\subsets\yeast_times100_subset_50.txt'); % 50 k=10
%X = load('X:\clustering\datasets\subsets\glass_subset_50.txt'); % 50 k=7
%X = load('X:\clustering\datasets\subsets\wdbc_subset_50.txt'); % 50 k=2
%X = load('..\..\datasets\a1.txt'); k=20; Dataset_name = 'a1'; ground_truth = '../../datasets/ground_truths/a1-ga-cb.txt';% 3000 points, 20 clust
%X = load('..\..\datasets\DIM032.txt'); k=16; Dataset_name = 'DIM032';ground_truth = '../../datasets/ground_truths/dim032-gt.txt';% 1024 points, 16 clust
%X = load('..\..\datasets\iris.txt'); % 150 points, 3 clust
%X = load('..\..\datasets\thyroid.txt'); % 215 points, 2 clust
%X = load('..\..\datasets\wine.txt'); % 178 points, 3 clust
%X = load('..\..\datasets\breast.txt'); % 699 points, 2 clust
%X = load('..\..\datasets\yeast_times100.txt'); % 10 clust
%X = load('..\..\datasets\glass.txt'); % 214 points, 7 clust
%X = load('..\..\datasets\wdbc.txt'); % 569 points, 2 clust
%X = load('X:\clustering\datasets\thyroid.txt'); % 5dim. 215 points 2 clust
%X = load('X:\clustering\datasets\wine.txt'); % 13 dim, 178 points, 3 clust
%X = load('X:\clustering\datasets\DIM032.txt'); % 1024 points,16 clust
%X = load('X:\clustering\datasets\DIM064.txt'); % 1024 points, 16 clust, 64 dim.
%X = load('X:\clustering\datasets\DIM128.txt'); % 1024 points,16 clust, 128 dim.
%X = load('X:\clustering\datasets\dim256.txt'); % 1024 points,16 clust, 256 dim.
%X = load('X:\clustering\datasets\bridge.txt'); % 16 dim, 4096 points, 256 clusters
%X = load('X:\clustering\datasets\housec8.txt');  % 3 dim 34k vect 256 clust
%X = load('X:\clustering\datasets\missa1.txt');  % 16 dim 6480 points 256 clust
%X = load('X:\clustering\datasets\yeast.txt');  % 1484 points, 10 clust
%X = load('.\MopsiLocations2012\mopsilocations2012Finland.txt');  
%X = load('..\..\datasets\g2_sets\g2-2-100.txt');  % Gaussian 2 clust.
%X = load('..\..\datasets\unbalance.txt'); k=8; Dataset_name = 'unbalance';ground_truth = '../../datasets/ground_truths/unbalance-gt.txt';% 8 clust


n = size(X,1);
d = size(X,2);

best_earth_movers = 0;  % some number

if CALCULATE_EPSILON == 1
W_total = 0;
for i=1:n-1
    for l=i+1:n
           W_total = W_total + (X(i,:)-X(l,:))*(X(i,:)-X(l,:))';
    end
end    
end


for repeats = 1:REPEATS  % 1:100  1:30

toc_last = toc;
    
% Initial centroids
for i = 1:k
    pass = 0;
    while (pass==0)
        pass = 1;
        C(i,:) = X(randi(n),:);
        for j = 1:i-1
            if C(j,:) == C(i,:) 
                pass=0;
            end
        end
    end

end

[partition,C] = kmeans(X,k,'start',C,'emptyaction','singleton');

for i=1:k
    number_of_points_in_cluster(i) = length(find(partition==i));
end


TSE = 0;
for i = 1:n
    TSE = TSE + ((X(i,:)-C(partition(i),:))*(X(i,:)-C(partition(i),:))');
end

for j= 1:k
    TSE_of_cluster(j) = 0;
end
for i = 1:n
    j = partition(i);
    TSE_of_cluster(j) = TSE_of_cluster(j) + ((X(i,:)-C(j,:))*(X(i,:)-C(j,:))');
end

%TSE_of_cluster

%while(1==1)
%end


rounds = 0;
point_changed_cluster = 1;
while point_changed_cluster == 1
%if (toc-toc_last)>60
%    break;
%end
rounds = rounds+1;
number_of_changes = 0;
point_changed_cluster=0;
for i=1:n
    tentative_point_no = i;  % i
%    while (number_of_points_in_cluster(partition(tentative_point_no))<2)
%        number_of_points_in_a_cluster_is_smaller_than_2 = 1
%        while(1==1) % endless loop
%        end
%        %tentative_point_no = tentative_point_no+1;
%    end
    for tentative_addition_cluster = 1:k

    if ((tentative_addition_cluster~=partition(tentative_point_no))&&(number_of_points_in_cluster(partition(tentative_point_no))>1))
        
    % deltaTSE = n1 * deltaTSEaddition + n2 * deltaTSEremoval
    %deltaTSEaddition = (number_of_points_in_cluster(tentative_addition_cluster)*1)/(number_of_points_in_cluster(tentative_addition_cluster)+1)*((C(tentative_addition_cluster,:)-X(tentative_point_no,:))*(C(tentative_addition_cluster,:)-X(tentative_point_no,:))');
    %deltaTSEremoval = -1/(number_of_points_in_cluster(partition(tentative_point_no))*(number_of_points_in_cluster(partition(tentative_addition_cluster))-1));
    %deltaTSE = (number_of_points_in_cluster(partition(tentative_addition_cluster))+1)*deltaTSEaddition+(number_of_points_in_cluster(partition(tentative_point_no))-1)*deltaTSEremoval;

    
    deltaTSEaddition = (number_of_points_in_cluster(tentative_addition_cluster)*1)/(number_of_points_in_cluster(tentative_addition_cluster)+1)*((C(tentative_addition_cluster,:)-X(tentative_point_no,:))*(C(tentative_addition_cluster,:)-X(tentative_point_no,:))');
%    deltaTSEremoval = -(number_of_points_in_cluster(partition(tentative_point_no))-1)/(number_of_points_in_cluster(partition(tentative_point_no)))*((number_of_points_in_cluster(partition(tentative_point_no))/(number_of_points_in_cluster(partition(tentative_point_no))-1))*C(partition(tentative_point_no),:)-(number_of_points_in_cluster(partition(tentative_point_no))-2)/(number_of_points_in_cluster(partition(tentative_point_no))-1)*X(tentative_point_no))*(number_of_points_in_cluster(partition(tentative_point_no))/(number_of_points_in_cluster(partition(tentative_point_no))-1)*C(partition(tentative_point_no),:)-(number_of_points_in_cluster(partition(tentative_point_no))-2)/(number_of_points_in_cluster(partition(tentative_point_no))-1)*X(tentative_point_no,:))';
    deltaTSEremoval = -(number_of_points_in_cluster(partition(tentative_point_no)))/(number_of_points_in_cluster(partition(tentative_point_no))-1)*((C(partition(tentative_point_no),:)-X(tentative_point_no,:))*(C(partition(tentative_point_no),:)-X(tentative_point_no,:))');

    deltaCOST = TSE_of_cluster(tentative_addition_cluster)+(number_of_points_in_cluster(tentative_addition_cluster)+1)*deltaTSEaddition-TSE_of_cluster(partition(tentative_point_no))+(number_of_points_in_cluster(partition(tentative_point_no))-1)*deltaTSEremoval;
    
    
    if (deltaCOST<0)
        %        C(tentative_addition_cluster,:) = C(tentative_addition_cluster,:)+1/(number_of_points_in_cluster(tentative_addition_cluster)+1)*(X(tentative_point_no,:)-C(tentative_addition_cluster,:));
        C(tentative_addition_cluster,:) = C(tentative_addition_cluster,:)*(number_of_points_in_cluster(tentative_addition_cluster))/(number_of_points_in_cluster(tentative_addition_cluster)+1)+1/(number_of_points_in_cluster(tentative_addition_cluster)+1)*(X(tentative_point_no,:));
        C(partition(tentative_point_no),:)= number_of_points_in_cluster(partition(tentative_point_no))/(number_of_points_in_cluster(partition(tentative_point_no))-1)*C(partition(tentative_point_no),:)-1/(number_of_points_in_cluster(partition(tentative_point_no))-1)*X(tentative_point_no,:);
        number_of_points_in_cluster(tentative_addition_cluster) = number_of_points_in_cluster(tentative_addition_cluster)+1;
        number_of_points_in_cluster(partition(tentative_point_no))=number_of_points_in_cluster(partition(tentative_point_no))-1;
        TSE_of_cluster(tentative_addition_cluster) = TSE_of_cluster(tentative_addition_cluster) + deltaTSEaddition;
        TSE_of_cluster(partition(tentative_point_no)) = TSE_of_cluster(partition(tentative_point_no)) + deltaTSEremoval;

        %saved_delta_TSE_removal(number_of_changes+1) = deltaTSEremoval;
        %saved_x(number_of_changes+1) = number_of_changes+1;
        
        %for jj= 1:k
        %    TSE_of_cluster(jj) = 0;
        %end
        %for ii = 1:n
        %    jj = partition(ii);
        %    TSE_of_cluster(jj) = TSE_of_cluster(jj) + ((X(ii,:)-C(jj,:))*(X(ii,:)-C(jj,:))');
        %end

        partition(tentative_point_no) = tentative_addition_cluster;
        point_changed_cluster=1;
        number_of_changes = number_of_changes +1;
        %deltaCOST
        break;  % to exit from for tentative_addition_cluster -loop
    end
    
    end % if tentative_addition_cluster~= ...
    end % for tentative_addition_cluster
    
end % for i
%rounds
%number_of_changes
end % while point changed cluster

partitions_array(repeats,:) = partition;

% Calculate CI-Index
% Partition has been just calculated
if ((strcmp(Dataset_name,'s1')) || (strcmp(Dataset_name,'s2')) || (strcmp(Dataset_name,'s3')) || (strcmp(Dataset_name,'s4')) || (strcmp(Dataset_name,'a1')) || (strcmp(Dataset_name,'DIM032'))|| (strcmp(Dataset_name,'unbalance')))
    gt = load(ground_truth);
    CCC(1:k,1:d) = 0;
    for j = 1:k
        CCC(j,:) = mean(X(find(partition(1:n)==j),:));
    end
    CI(repeats) = centroidIndex3(CCC,gt);
end



% calculate the earth mover's distance
% new formula
for i = 1:k
    cluster_size(i) = 0;
end
for i = 1:n
    cluster_size(partition(i)) = cluster_size(partition(i)) + 1;
end
earth_movers = 0;
for j = 1:k
    %cluster_size(i)
    earth_movers = earth_movers + max(cluster_size(j)-ceil(n/k),floor(n/k)-cluster_size(j));
end
earth_movers_array(repeats) = earth_movers;
    if (earth_movers < best_earth_movers)||(repeats==1)
        best_earth_movers = earth_movers;
        best_partition = partition;
    end
    
   
if CALCULATE_EPSILON == 1
value_of_cut(repeats) = 0;
for i=1:n-1
    for l=i+1:n
        if (partition(i)~=partition(l))
           value_of_cut(repeats) = value_of_cut(repeats) + (X(i,:)-X(l,:))*(X(i,:)-X(l,:))';
        end
    end
end    
end


for j = 1:k
C(j,:) = mean(X(find(partition==j),:));
end
MSE_runs(repeats) = 0;
for i = 1:n
    MSE_runs(repeats) = MSE_runs(repeats) + ((X(i,:)-C(partition(i),:))*(X(i,:)-C(partition(i),:))')/n;
end


CostScut = 0;
for j=1:k
CostScut = CostScut + cluster_size(j)*TSE_of_cluster(j);
end
Cost_Scut_runs(repeats) = CostScut;
if repeats==1 
    bestCostScut= CostScut;
    bestCostScutPartition = partition;
elseif CostScut<bestCostScut
    bestCostScut=CostScut;
    bestCostScutPartition = partition;
end



repeats
end %repeats


partition = best_partition;



for j = 1:k
C(j,:) = mean(X(find(partition==j),:));
end
MSE = 0;
for i = 1:n
    MSE = MSE + ((X(i,:)-C(partition(i),:))*(X(i,:)-C(partition(i),:))')/n;
end

if SAVEPARTITION == 1
    dlmwrite('partition.txt',bestCostScutPartition,'delimiter',' ');
end

MSE_of_best_earth_movers = MSE

MSE_best = min(MSE_runs)

mean_MSE = mean(MSE_runs)

best_earth_movers = best_earth_movers
mean_earth_movers = mean(earth_movers_array)

mean_Scut = mean(Cost_Scut_runs)
best_Scut = min(Cost_Scut_runs)

if CALCULATE_EPSILON == 1
W_total
mean_value_of_cut = mean(value_of_cut)
alpha_k = 1-1/k;
epsilon_k = (W_total-mean(value_of_cut))/(W_total-1/alpha_k*mean(value_of_cut))
end
partition = bestCostScutPartition;

if ((strcmp(Dataset_name,'s1')) || (strcmp(Dataset_name,'s2')) || (strcmp(Dataset_name,'s3')) || (strcmp(Dataset_name,'s4'))|| (strcmp(Dataset_name,'a1')) || (strcmp(Dataset_name,'DIM032'))|| (strcmp(Dataset_name,'unbalance')))
mean_CI_index = mean(CI)
end

mean_time = toc/repeats

figure   
plot(X(find(partition==1),1),X(find(partition==1),2),'r+');
if k>1
    hold on
    plot(X(find(partition==2),1),X(find(partition==2),2),'bO');
end
if k>2
    hold on
    plot(X(find(partition==3),1),X(find(partition==3),2),'r.');
end
if k>3
    hold on
    plot(X(find(partition==4),1),X(find(partition==4),2),'b.');
end
if k>4
    hold on
    plot(X(find(partition==5),1),X(find(partition==5),2),'r+');
end
if k>5
    hold on
    plot(X(find(partition==6),1),X(find(partition==6),2),'bO');
end
if k>6
    hold on
    plot(X(find(partition==7),1),X(find(partition==7),2),'gO');
end
if k>7
    hold on
    plot(X(find(partition==8),1),X(find(partition==8),2),'g.');
end
if k>8
    hold on
    plot(X(find(partition==9),1),X(find(partition==9),2),'r.');
end
if k>9
    hold on
    plot(X(find(partition==10),1),X(find(partition==10),2),'b.');
end
if k>10
    hold on
    plot(X(find(partition==11),1),X(find(partition==11),2),'g+');
end
if k>11
    hold on
    plot(X(find(partition==12),1),X(find(partition==12),2),'gO');
end
if k>12
    hold on
    plot(X(find(partition==13),1),X(find(partition==13),2),'g+');
end
if k>13
    hold on
    plot(X(find(partition==14),1),X(find(partition==14),2),'g.');
end
if k>14
    hold on
    plot(X(find(partition==15),1),X(find(partition==15),2),'g.');
end

end % function
